@@ -39,15 +39,24 @@
 # define EXIT_FAILURE 1
 #endif
 
+#if defined HAVE_WCTYPE_H && defined HAVE_WCHAR_H && defined HAVE_MBRTOWC
+/* We can handle multibyte string.  */
+# include <wchar.h>
+# include <wctype.h>
+# define MBS_SUPPORT
+#endif
+
 #ifdef HAVE_SYS_TYPES_H
 # include <sys/types.h>
 #endif
-
 #include <obstack.h>
-
+#include "basicdefs.h"
 #include "utils.h"
 #include "sed.h"
 
+extern flagT no_default_output;
+extern flagT POSIXLY_CORRECT;
+
 
 #define YMAP_LENGTH		256 /*XXX shouldn't this be (UCHAR_MAX+1)?*/
 #define VECTOR_ALLOC_INCREMENT	40
@@ -96,18 +105,28 @@ struct sed_label {
 };
 
 struct special_files {
-  struct output outf;
+  char *name;
   FILE **pfp;
 };
 
 FILE *my_stdin, *my_stdout, *my_stderr;
 struct special_files special_files[] = {
-  { { "/dev/stdin", false, NULL, NULL }, &my_stdin },
-  { { "/dev/stdout", false, NULL, NULL }, &my_stdout },
-  { { "/dev/stderr", false, NULL, NULL }, &my_stderr },
-  { { NULL, false, NULL, NULL }, NULL }
+  { "/dev/stdin", &my_stdin },
+  { "/dev/stdout", &my_stdout },
+  { "/dev/stderr", &my_stderr },
+  { NULL, NULL }
 };
 
+/* This structure tracks files opened by the `w' and `s///w' commands
+   so that they may all be closed cleanly at normal program termination.
+   Those marked as `special' are not closed. */
+struct fp_list {
+    char *name;
+    int special;
+    FILE *fp;
+    struct fp_list *link;
+  };
+
 
 /* Where we are in the processing of the input. */
 static struct prog_info prog;
@@ -120,7 +139,7 @@ static struct sed_label *labels = NULL;
 
 /* We wish to detect #n magic only in the first input argument;
    this flag tracks when we have consumed the first file of input. */
-static bool first_script = true;
+static flagT first_script = TRUE;
 
 /* Allow for scripts like "sed -e 'i\' -e foo": */
 static struct buffer *pending_text = NULL;
@@ -135,30 +154,28 @@ static struct obstack obs;
 
 /* Various error messages we may want to print */
 static const char errors[] =
-  "multiple `!'s\0"
-  "unexpected `,'\0"
-  "cannot use +N or ~N as first address\0"
-  "unmatched `{'\0"
-  "unexpected `}'\0"
-  "extra characters after command\0"
-  "expected \\ after `a', `c' or `i'\0"
+  "Multiple `!'s\0"
+  "Unexpected `,'\0"
+  "Cannot use +N or ~N as first address\0"
+  "Unmatched `{'\0"
+  "Unexpected `}'\0"
+  "Extra characters after command\0"
+  "Expected \\ after `a', `c' or `i'\0"
   "`}' doesn't want any addresses\0"
   ": doesn't want any addresses\0"
-  "comments don't accept any addresses\0"
-  "missing command\0"
-  "command only uses one address\0"
-  "unterminated address regex\0"
-  "unterminated `s' command\0"
-  "unterminated `y' command\0"
-  "unknown option to `s'\0"
+  "Comments don't accept any addresses\0"
+  "Missing command\0"
+  "Command only uses one address\0"
+  "Unterminated address regex\0"
+  "Unterminated `s' command\0"
+  "Unterminated `y' command\0"
+  "Unknown option to `s'\0"
   "multiple `p' options to `s' command\0"
   "multiple `g' options to `s' command\0"
   "multiple number options to `s' command\0"
   "number option to `s' command may not be zero\0"
-  "strings for `y' command are different lengths\0"
-  "delimiter character is not a single-byte character\0"
-  "expected newer version of sed\0"
-  "invalid usage of line address 0";
+  "strings for y command are different lengths\0"
+  "expected newer version of sed";
 
 #define BAD_BANG (errors)
 #define BAD_COMMA (BAD_BANG + sizeof(N_("Multiple `!'s")))
@@ -181,13 +198,11 @@ static const char errors[] =
 #define EXCESS_N_OPT (EXCESS_G_OPT + sizeof(N_("multiple `g' options to `s' command")))
 #define ZERO_N_OPT (EXCESS_N_OPT + sizeof(N_("multiple number options to `s' command")))
 #define Y_CMD_LEN (ZERO_N_OPT + sizeof(N_("number option to `s' command may not be zero")))
-#define BAD_DELIM (Y_CMD_LEN + sizeof(N_("strings for `y' command are different lengths")))
-#define ANCIENT_VERSION (BAD_DELIM + sizeof(N_("delimiter character is not a single-byte character")))
-#define INVALID_LINE_0 (ANCIENT_VERSION + sizeof(N_("expected newer version of sed")))
-#define END_ERRORS (INVALID_LINE_0 + sizeof(N_("invalid usage of line address 0")))
+#define ANCIENT_VERSION (Y_CMD_LEN + sizeof(N_("strings for y command are different lengths")))
+#define END_ERRORS (ANCIENT_VERSION + sizeof(N_("expected newer version of sed")))
 
-static struct output *file_read = NULL;
-static struct output *file_write = NULL;
+static struct fp_list *file_read = NULL;
+static struct fp_list *file_write = NULL;
 
 
 /* Read the next character from the program.  Return EOF if there isn't
@@ -273,34 +288,6 @@ add_then_next(b, ch)
   return inchar();
 }
 
-/* Add a byte to the multibyte character represented by the state
-   CUR_STAT, and answer its length if a character is completed,
-   or -2 if it is yet to be completed.  */
-static int brlen P_((int ch, MBSTATE_T *cur_stat));
-static int brlen (ch, cur_stat)
-     int ch;
-     MBSTATE_T *cur_stat;
-{
-#ifdef MBS_SUPPORT
-  char c = ch;
-
-  /* Calling mbrtowc with the first parameter = NULL is the same
-     as mbrlen.  */
-  int result = mbrtowc(NULL, &c, 1, cur_stat);
-
-  /* An invalid sequence is treated like a singlebyte character. */
-  if (result == -1)
-    {
-      memset (cur_stat, 0, sizeof (MBSTATE_T));
-      return 1;
-    }
-
-  return result;
-#else
-  return 1;
-#endif
-}
-
 static char * convert_number P_((char *, char *, const char *, int, int, int));
 static char *
 convert_number(result, buf, bufend, base, maxdigits, default_char)
@@ -362,28 +349,29 @@ read_filename()
     {
 #if 0 /*XXX ZZZ 1998-09-12 kpp: added, then had second thoughts*/
       if (!POSIXLY_CORRECT)
-	if (ch == ';' || ch == '#')
+	if (ch == ';' || ch == CLOSE_BRACE || ch == '#')
 	  {
 	    savchar(ch);
 	    break;
 	  }
 #endif
-      ch = add_then_next(b, ch);
+      add1_buffer(b, ch);
+      ch = inchar();
     }
   add1_buffer(b, '\0');
   return b;
 }
 
-static struct output *get_openfile P_((struct output **file_ptrs, char *mode, bool fail));
-static struct output *
+static FILE *get_openfile P_((struct fp_list **file_ptrs, char *mode, flagT fail));
+static FILE *
 get_openfile(file_ptrs, mode, fail)
-     struct output **file_ptrs;
+     struct fp_list **file_ptrs;
      char *mode;
-     bool fail;
+     flagT fail;
 {
   struct buffer *b;
   char *file_name;
-  struct output *p;
+  struct fp_list *p;
   int is_stderr;
 
   b = read_filename();
@@ -392,34 +380,36 @@ get_openfile(file_ptrs, mode, fail)
     if (strcmp(p->name, file_name) == 0)
       break;
 
-  if (!POSIXLY_CORRECT)
-    {
-      /* Check whether it is a special file (stdin, stdout or stderr) */
-      struct special_files *special = special_files;
-		  
-      /* std* sometimes are not constants, so they
-         cannot be used in the initializer for special_files */
-      my_stdin = stdin; my_stdout = stdout; my_stderr = stderr;
-      for (special = special_files; special->outf.name; special++)
-        if (strcmp(special->outf.name, file_name) == 0)
-          {
-	    special->outf.fp = *special->pfp;
-	    free_buffer (b);
-	    return &special->outf;
-          }
-    }
-
   if (!p)
     {
-      p = OB_MALLOC(&obs, 1, struct output);
+      FILE *fp = NULL;
+      if (!POSIXLY_CORRECT)
+	{
+	  /* Check whether it is a special file (stdin, stdout or stderr) */
+	  struct special_files *special = special_files;
+		  
+	  /* std* sometimes are not constants, so they
+	     cannot be used in the initializer for special_files */
+	  my_stdin = stdin; my_stdout = stdout; my_stderr = stderr;
+	  for (special = special_files; special->name; special++)
+	    if (strcmp(special->name, file_name) == 0)
+	      {
+		fp = *special->pfp;
+		break;
+	      }
+	}
+
+      p = OB_MALLOC(&obs, 1, struct fp_list);
       p->name = ck_strdup(file_name);
-      p->fp = ck_fopen(p->name, mode, fail);
-      p->missing_newline = false;
+      p->special = fp != NULL;
+      if (!fp)
+	fp = ck_fopen(p->name, mode, fail);
+      p->fp = fp;
       p->link = *file_ptrs;
       *file_ptrs = p;
     }
   free_buffer(b);
-  return p;
+  return p->fp;
 }
 
 
@@ -441,161 +431,99 @@ next_cmd_entry(vectorp)
   cmd = v->v + v->v_length;
   cmd->a1 = NULL;
   cmd->a2 = NULL;
-  cmd->a1_matched = false;
-  cmd->addr_bang = false;
+  cmd->a1_matched = FALSE;
+  cmd->addr_bang = FALSE;
   cmd->cmd = '\0';	/* something invalid, to catch bugs early */
 
   *vectorp  = v;
   return cmd;
 }
 
-static int snarf_char_class P_((struct buffer *b, MBSTATE_T *cur_stat));
+static int snarf_char_class P_((struct buffer *b));
 static int
-snarf_char_class(b, cur_stat)
+snarf_char_class(b)
   struct buffer *b;
-  MBSTATE_T *cur_stat;
 {
   int ch;
-  int state = 0;
-  int delim;
-  bool pending_mb = 0;
-  
+
   ch = inchar();
   if (ch == '^')
     ch = add_then_next(b, ch);
-  else if (ch == CLOSE_BRACKET)
+  if (ch == CLOSE_BRACKET)
     ch = add_then_next(b, ch);
-  
-  /* States are:
-	0 outside a collation element, character class or collation class
-	1 after the bracket
-	2 after the opening ./:/=
-	3 after the closing ./:/= */
-
-  for (;; ch = add_then_next (b, ch))
+  while (ch != EOF && ch != '\n' && ch != CLOSE_BRACKET)
     {
-      pending_mb = brlen (ch, cur_stat) != 1;
-
-      switch (ch)
+      if (ch == OPEN_BRACKET)
 	{
-	case EOF:
-	case '\n':
-	  return ch;
+	  int prev;
+	  int delim = ch = add_then_next(b, ch);
 
-	case '.':
-	case ':':
-	case '=':
-	  if (pending_mb)
-	    continue;
-
-	  if (state == 1)
+	  if (delim != '.'  &&  delim != ':'  &&  delim != '=')
+	    continue; /* bypass the add_then_next() call at bottom of loop */
+	  for (prev=ch=add_then_next(b, ch);
+		  !(ch==CLOSE_BRACKET && prev==delim); ch=add_then_next(b, ch))
 	    {
-	      delim = ch;
-	      state++;
+	      if (ch == EOF || ch == '\n')
+		return ch;
+	      prev = ch;
 	    }
-	  else if (ch == delim && state == 2)
-	    state++;
-	  else
-	    break;
-
-	  continue;
-
-	case OPEN_BRACKET:
-	  if (pending_mb)
-	    continue;
-
-	  state++;
-	  continue;
-
-	case CLOSE_BRACKET:
-	  if (pending_mb)
-	    continue;
-
-	  if (state == 0)
-	    return ch;
-	  else if (state == 3)
-	    state = 0;
-
-	  break;
-
+	}
 #ifndef REG_PERL
-	case '\\':
-	  if (pending_mb)
-	    continue;
-
+      else if (ch == '\\')
+	{
 	  ch = inchar();
 	  if (ch == EOF)
-	    return EOF;
-
+	    break;
 	  if (ch != 'n' && ch != '\n')
-	    add1_buffer(b, '\\');
-	  else
-	    ch = '\n';
-
-	  break;
-#endif
-
-	default:
-	  break;
+	    {
+	      add1_buffer(b, '\\');
+	      continue; /* bypass the add_then_next() call at bottom of loop */
+	    }
+	  ch = '\n';
 	}
-
-      /* Getting a character different from .=: whilst in state 1
-         goes back to state 0, getting a character different from ]
-         whilst in state 3 goes back to state 2.  */
-      state &= ~1;
+#endif
+      ch = add_then_next(b, ch);
     }
+
+  return ch;
 }
 
-static struct buffer *match_slash P_((int slash, bool regex));
+static struct buffer *match_slash P_((int slash, flagT regex, flagT keep_back));
 static struct buffer *
-match_slash(slash, regex)
+match_slash(slash, regex, keep_backwhack)
   int slash;
-  bool regex;
+  flagT regex;
+  flagT keep_backwhack;
 {
   struct buffer *b;
   int ch;
-  bool pending_mb = false;
-  MBSTATE_T cur_stat;
-
-  memset (&cur_stat, 0, sizeof (MBSTATE_T));
-
-  if (brlen (slash, &cur_stat) == -2)
-    bad_prog (BAD_DELIM);
-
-  memset (&cur_stat, 0, sizeof (MBSTATE_T));
 
   b = init_buffer();
-  while ((ch = inchar()) != EOF && ch != '\n')
+  while ((ch = inchar()) != EOF && ch != '\n' && ch != slash)
     {
-      pending_mb = brlen (ch, &cur_stat) != 1;
-
-      if (!pending_mb)
+      if (ch == '\\')
 	{
-	  if (ch == slash)
-	    return b;
-	  else if (ch == '\\')
-	    {
-	      ch = inchar();
-	      if (ch == EOF)
-	        break;
+	  ch = inchar();
+	  if (ch == EOF)
+	    break;
 #ifndef REG_PERL
-	      else if (ch == 'n' && regex)
-	        ch = '\n';
+	  else if (ch == 'n' && regex)
+	    ch = '\n';
 #endif
-	      else if (ch != '\n' && ch != slash)
-	        add1_buffer(b, '\\');
-	    }
-          else if (ch == OPEN_BRACKET && regex)
-	    {
-	      add1_buffer(b, ch);
-	      ch = snarf_char_class(b, &cur_stat);
-	      if (ch != CLOSE_BRACKET)
-	        break;
-	    }
+	  else if (ch != '\n' && (ch != slash || keep_backwhack))
+	    add1_buffer(b, '\\');
+	}
+      else if (ch == OPEN_BRACKET && regex)
+	{
+	  add1_buffer(b, ch);
+	  ch = snarf_char_class(b);
+	  if (ch != CLOSE_BRACKET)
+	    break;
 	}
-
       add1_buffer(b, ch);
     }
+  if (ch == slash)
+    return b;
 
   if (ch == '\n')
     savchar(ch);	/* for proper line number in error report */
@@ -603,19 +531,19 @@ match_slash(slash, regex)
   return NULL;
 }
 
-static bool mark_subst_opts P_((struct subst *cmd));
-static bool
+static flagT mark_subst_opts P_((struct subst *cmd));
+static flagT
 mark_subst_opts(cmd)
   struct subst *cmd;
 {
   int flags = 0;
   int ch;
 
-  cmd->global = false;
-  cmd->print = false;
-  cmd->eval = false;
-  cmd->numb = 0;
-  cmd->outf = NULL;
+  cmd->global = FALSE;
+  cmd->print = FALSE;
+  cmd->eval = FALSE;
+  cmd->numb = FALSE;
+  cmd->fp = NULL;
 
   for (;;)
     switch ( (ch = in_nonblank()) )
@@ -645,23 +573,23 @@ mark_subst_opts(cmd)
 	break;
 
       case 'e':
-	cmd->eval = true;
+	cmd->eval = TRUE;
 	break;
 
       case 'p':
 	if (cmd->print)
 	  bad_prog(_(EXCESS_P_OPT));
-	cmd->print |= (1 << cmd->eval); /* 1=before eval, 2=after */
+	cmd->print |= (TRUE << cmd->eval); /* 1=before eval, 2=after */
 	break;
 
       case 'g':
 	if (cmd->global)
 	  bad_prog(_(EXCESS_G_OPT));
-	cmd->global = true;
+	cmd->global = TRUE;
 	break;
 
       case 'w':
-	cmd->outf = get_openfile(&file_write, "w", true);
+	cmd->fp = get_openfile(&file_write, "w", TRUE);
 	return flags;
 
       case '0': case '1': case '2': case '3': case '4':
@@ -708,8 +636,10 @@ read_label()
 
   while (ch != EOF && ch != '\n'
 	 && !ISBLANK(ch) && ch != ';' && ch != CLOSE_BRACE && ch != '#')
-    ch = add_then_next (b, ch);
-
+    {
+      add1_buffer(b, ch);
+      ch = inchar();
+    }
   savchar(ch);
   add1_buffer(b, '\0');
   ret = ck_strdup(get_buffer(b));
@@ -734,7 +664,7 @@ setup_label(list, idx, name, err_info)
   ret->v_index = idx;
   ret->name = name;
   if (err_info)
-    MEMCPY(&ret->err_info, err_info, sizeof (ret->err_info));
+    MEMCPY(&ret->err_info, err_info, sizeof ret->err_info);
   ret->next = list;
   return ret;
 }
@@ -788,13 +718,13 @@ setup_replacement(sub, text, length)
   char *base;
   char *p;
   char *text_end;
-  enum replacement_types repl_type = REPL_ASIS, save_type = REPL_ASIS;
+  enum replacement_types repl_type = repl_asis, save_type = repl_asis;
   struct replacement root;
   struct replacement *tail;
 
   sub->max_id = 0;
   base = MEMDUP(text, length, char);
-  length = normalize_text(base, length, false);
+  length = normalize_text(base, length);
 
   text_end = base + length;
   tail = &root;
@@ -809,10 +739,9 @@ setup_replacement(sub, text, length)
 
 	  repl_type = save_type;
 
-	  /* Skip the backslash and look for a numeric back-reference,
-	     or a case-munging escape if not in POSIXLY_CORRECT mode: */
+	  /* Skip the backslash and look for a numeric back-reference: */
 	  ++p;
-	  if (p<text_end && (!POSIXLY_CORRECT || ISDIGIT (*p)))
+	  if (p<text_end)
 	    switch (*p)
 	      {
 	      case '0': case '1': case '2': case '3': case '4': 
@@ -823,28 +752,28 @@ setup_replacement(sub, text, length)
 		break;
 
 	      case 'L':
-		repl_type = REPL_LOWERCASE;
-		save_type = REPL_LOWERCASE;
+		repl_type = repl_lowercase;
+		save_type = repl_lowercase;
 		break;
 
 	      case 'U':
-		repl_type = REPL_UPPERCASE;
-		save_type = REPL_UPPERCASE;
+		repl_type = repl_uppercase;
+		save_type = repl_uppercase;
 		break;
 		
 	      case 'E':
-		repl_type = REPL_ASIS;
-		save_type = REPL_ASIS;
+		repl_type = repl_asis;
+		save_type = repl_asis;
 		break;
 
 	      case 'l':
 		save_type = repl_type;
-		repl_type |= REPL_LOWERCASE_FIRST;
+		repl_type |= repl_lowercase_first;
 		break;
 
 	      case 'u':
 		save_type = repl_type;
-		repl_type |= REPL_UPPERCASE_FIRST;
+		repl_type |= repl_uppercase_first;
 		break;
 		
 	      default:
@@ -896,34 +825,42 @@ read_text(buf, leadin_ch)
 
   if (leadin_ch == EOF)
     return;
-
+#ifndef NO_INPUT_INDENT
   if (leadin_ch != '\n')
     add1_buffer(pending_text, leadin_ch);
-
   ch = inchar();
-  while (ch != EOF && ch != '\n')
+#else /*NO_INPUT_INDENT*/
+  if (leadin_ch == '\n')
+    ch = in_nonblank();
+  else
+    {
+      add1_buffer(pending_text, leadin_ch);
+      ch = inchar();
+    }
+#endif /*NO_INPUT_INDENT*/
+
+  while (ch!=EOF && ch!='\n')
     {
       if (ch == '\\')
-	{
-	  ch = inchar();
-	  if (ch != EOF)
-	    add1_buffer (pending_text, '\\');
-	}
-
+	ch = inchar();
       if (ch == EOF)
 	{
-	  add1_buffer (pending_text, '\n');
+	  add1_buffer(pending_text, '\n');
 	  return;
 	}
-
-      ch = add_then_next (pending_text, ch);
+      add1_buffer(pending_text, ch);
+#ifdef NO_INPUT_INDENT
+      if (ch == '\n')
+	ch = in_nonblank();
+      else
+#endif /*NO_INPUT_INDENT*/
+	ch = inchar();
     }
-
   add1_buffer(pending_text, '\n');
+
   if (!buf)
     buf = old_text_buf;
-  buf->text_length = normalize_text (get_buffer (pending_text),
-				     size_buffer (pending_text), true);
+  buf->text_length = size_buffer(pending_text);
   buf->text = MEMDUP(get_buffer(pending_text), buf->text_length, char);
   free_buffer(pending_text);
   pending_text = NULL;
@@ -934,13 +871,13 @@ read_text(buf, leadin_ch)
    return non-zero and store the resulting address in `*addr'.
    If the input doesn't look like an address read nothing
    and return zero.  */
-static bool compile_address P_((struct addr *addr, int ch));
-static bool
+static flagT compile_address P_((struct addr *addr, int ch));
+static flagT
 compile_address(addr, ch)
   struct addr *addr;
   int ch;
 {
-  addr->addr_type = ADDR_IS_NULL;
+  addr->addr_type = addr_is_null;
   addr->addr_step = 0;
   addr->addr_number = ~(countT)0;  /* extremely unlikely to ever match */
   addr->addr_regex = NULL;
@@ -949,10 +886,10 @@ compile_address(addr, ch)
     {
       int flags = 0;
       struct buffer *b;
-      addr->addr_type = ADDR_IS_REGEX;
+      addr->addr_type = addr_is_regex;
       if (ch == '\\')
 	ch = inchar();
-      if ( !(b = match_slash(ch, true)) )
+      if ( !(b = match_slash(ch, TRUE, TRUE)) )
 	bad_prog(_(UNTERM_ADDR_RE));
 
       for(;;)
@@ -984,14 +921,14 @@ compile_address(addr, ch)
 	      savchar (ch);
 	      addr->addr_regex = compile_regex (b, flags, 0);
 	      free_buffer(b);
-	      return true;
+	      return TRUE;
 	    }
 	}
     }
   else if (ISDIGIT(ch))
     {
       addr->addr_number = in_integer(ch);
-      addr->addr_type = ADDR_IS_NUM;
+      addr->addr_type = addr_is_num;
       ch = in_nonblank();
       if (ch != '~')
 	{
@@ -1003,7 +940,7 @@ compile_address(addr, ch)
 	  if (step > 0)
 	    {
 	      addr->addr_step = step;
-	      addr->addr_type = ADDR_IS_NUM_MOD;
+	      addr->addr_type = addr_is_num_mod;
 	    }
 	}
     }
@@ -1011,20 +948,20 @@ compile_address(addr, ch)
     {
       addr->addr_step = in_integer(in_nonblank());
       if (addr->addr_step==0)
-	; /* default to ADDR_IS_NULL; forces matching to stop on next line */
+	; /* default to addr_is_null; forces matching to stop on next line */
       else if (ch == '+')
-	addr->addr_type = ADDR_IS_STEP;
+	addr->addr_type = addr_is_step;
       else
-	addr->addr_type = ADDR_IS_STEP_MOD;
+	addr->addr_type = addr_is_step_mod;
     }
   else if (ch == '$')
     {
-      addr->addr_type = ADDR_IS_LAST;
+      addr->addr_type = addr_is_last;
     }
   else
-    return false;
+    return FALSE;
 
-  return true;
+  return TRUE;
 }
 
 /* Read a program (or a subprogram within `{' `}' pairs) in and store
@@ -1062,8 +999,8 @@ compile_program(vector)
       cur_cmd = next_cmd_entry(&vector);
       if (compile_address(&a, ch))
 	{
-	  if (a.addr_type == ADDR_IS_STEP
-	      || a.addr_type == ADDR_IS_STEP_MOD)
+	  if (a.addr_type == addr_is_step
+	      || a.addr_type == addr_is_step_mod)
 	    bad_prog(_(BAD_PLUS));
 
 	  cur_cmd->a1 = MEMDUP(&a, 1, struct addr);
@@ -1072,25 +1009,19 @@ compile_program(vector)
 	    {
 	      if (!compile_address(&a, in_nonblank()))
 		bad_prog(_(BAD_COMMA));
-	      if (a.addr_type == ADDR_IS_NUM)
-		a.addr_type = ADDR_IS_NUM2;
+	      if (a.addr_type == addr_is_num)
+		a.addr_type = addr_is_num2;
 	      cur_cmd->a2 = MEMDUP(&a, 1, struct addr);
 	      ch = in_nonblank();
-
-	      if ((cur_cmd->a1->addr_type == ADDR_IS_NUM
-	           || cur_cmd->a1->addr_type == ADDR_IS_NUM_MOD)
-	          && cur_cmd->a1->addr_number == 0)
-	        cur_cmd->a1_matched = true;
+	      if ((cur_cmd->a1->addr_type == addr_is_num
+		   || cur_cmd->a1->addr_type == addr_is_num_mod)
+	      	  && cur_cmd->a1->addr_number == 0)
+		cur_cmd->a1_matched = TRUE;
 	    }
-
-	  if (cur_cmd->a1->addr_type == ADDR_IS_NUM
-	      && cur_cmd->a1->addr_number == 0
-	      && (!cur_cmd->a2 || cur_cmd->a2->addr_type != ADDR_IS_REGEX))
-	    bad_prog(_(INVALID_LINE_0));
 	}
       if (ch == '!')
 	{
-	  cur_cmd->addr_bang = true;
+	  cur_cmd->addr_bang = TRUE;
 	  ch = in_nonblank();
 	  if (ch == '!')
 	    bad_prog(_(BAD_BANG));
@@ -1106,26 +1037,26 @@ compile_program(vector)
 	  if (ch=='n' && first_script && cur_input.line < 2)
 	    if (   (prog.base && prog.cur==2+prog.base)
 		|| (prog.file && !prog.base && 2==ftell(prog.file)))
-	      no_default_output = true;
+	      no_default_output = TRUE;
 	  while (ch != EOF && ch != '\n')
 	    ch = inchar();
 	  continue;	/* restart the for (;;) loop */
 
 	case 'v':
-	  /* This is an extension.  Programs needing GNU sed might start
-	   * with a `v' command so that other seds will stop.
-	   * We compare the version ignore it.
-	   */
-	  {
-	    char *version = read_label ();
-	    char *compared_version;
-	    compared_version = (*version == '\0') ? "4.0" : version;
-	    if (strverscmp (compared_version, SED_FEATURE_VERSION) > 0)
-	      bad_prog(_(ANCIENT_VERSION));
+          /* This is an extension.  Programs needing GNU sed might start
+           * with a `v' command so that other seds will stop.
+           * We compare the version ignore it.
+           */
+          {
+            char *version = read_label ();
+            char *compared_version;
+            compared_version = (*version == '\0') ? "4.0" : version;
+            if (strverscmp (compared_version, SED_FEATURE_VERSION) > 0)
+              bad_prog(_(ANCIENT_VERSION));
 
-	    free (version);
+            free (version);
 	    POSIXLY_CORRECT = 0;
-	  }
+          }
 	  continue;
 
 	case '{':
@@ -1252,12 +1183,12 @@ compile_program(vector)
 	  break;
 
         case 'R':
-	  cur_cmd->x.fp = get_openfile(&file_read, "r", false)->fp;
+	  cur_cmd->x.fp = get_openfile(&file_read, "r", FALSE);
 	  break;
 
 	case 'w':
         case 'W':
-	  cur_cmd->x.outf = get_openfile(&file_write, "w", true);
+	  cur_cmd->x.fp = get_openfile(&file_write, "w", TRUE);
 	  break;
 
 	case 's':
@@ -1267,9 +1198,9 @@ compile_program(vector)
 	    int slash;
 
 	    slash = inchar();
-	    if ( !(b  = match_slash(slash, true)) )
+	    if ( !(b  = match_slash(slash, TRUE, TRUE)) )
 	      bad_prog(_(UNTERM_S_CMD));
-	    if ( !(b2 = match_slash(slash, false)) )
+	    if ( !(b2 = match_slash(slash, FALSE, TRUE)) )
 	      bad_prog(_(UNTERM_S_CMD));
 
 	    cur_cmd->x.cmd_subst = OB_MALLOC(&obs, 1, struct subst);
@@ -1286,33 +1217,37 @@ compile_program(vector)
 
 	case 'y':
 	  {
-	    size_t len, dest_len;
+	    unsigned char *ustring;
+	    size_t len;
 	    int slash;
-	    struct buffer *b2;
-            char *src_buf, *dest_buf;
-  
+
+#if defined MBS_SUPPORT && !defined (REG_PERL)
+	    if (MB_CUR_MAX == 1)
+#endif
+	      {
+	        ustring = OB_MALLOC(&obs, YMAP_LENGTH, unsigned char);
+	        for (len = 0; len < YMAP_LENGTH; len++)
+	          ustring[len] = len;
+	        cur_cmd->x.translate = ustring;
+	      }
+
 	    slash = inchar();
-	    if ( !(b = match_slash(slash, true)) )
+	    if ( !(b = match_slash(slash, FALSE, FALSE)) )
 	      bad_prog(_(UNTERM_Y_CMD));
-            src_buf = get_buffer(b);
-	    len = normalize_text(src_buf, size_buffer (b), true);
 
-            if ( !(b2 = match_slash(slash, true)) )
- 	      bad_prog(_(UNTERM_Y_CMD));
-            dest_buf = get_buffer(b2);
-	    dest_len = normalize_text(dest_buf, size_buffer (b2), true);
-
-#ifdef MBS_SUPPORT
+#if defined MBS_SUPPORT && !defined (REG_PERL)
             if (MB_CUR_MAX > 1)
 	      {
-                int i, j, idx, src_char_num;
+                int i, j, idx, src_char_num, len = size_buffer(b);
                 size_t *src_lens = MALLOC(len, size_t);
-                char **trans_pairs;
+                char *src_buf, *dest_buf, **trans_pairs;
                 size_t mbclen;
                 mbstate_t cur_stat;
+	        struct buffer *b2;
 
                 /* Enumerate how many character the source buffer has.  */
                 memset(&cur_stat, 0, sizeof(mbstate_t));
+                src_buf = get_buffer(b);
                 for (i = 0, j = 0; i < len;)
                   {
                     mbclen = mbrlen(src_buf + i, len - i, &cur_stat);
@@ -1327,7 +1262,11 @@ compile_program(vector)
                 src_char_num = j;
 
                 memset(&cur_stat, 0, sizeof(mbstate_t));
+                if ( !(b2 = match_slash(slash, FALSE, FALSE)) )
+ 		  bad_prog(_(UNTERM_Y_CMD));
+                dest_buf = get_buffer(b2);
                 idx = 0;
+                len = size_buffer(b2);
 
                 /* trans_pairs = {src(0), dest(0), src(1), dest(1), ..., NULL}
                      src(i) : pointer to i-th source character.
@@ -1363,30 +1302,33 @@ compile_program(vector)
                 trans_pairs[2 * i] = NULL;
                 if (idx != len)
                   bad_prog(_(Y_CMD_LEN));
+                free_buffer(b);
+                free_buffer(b2);
               }
             else
 #endif
               {
-	        char *translate = OB_MALLOC(&obs, YMAP_LENGTH, char);
-                unsigned char *ustring = CAST(unsigned char *)src_buf;
+                ustring = CAST(unsigned char *)get_buffer(b);
+                for (len = size_buffer(b); len; --len)
+                  {
+                    ch = inchar();
+                    if (ch == slash)
+                      bad_prog(_(Y_CMD_LEN));
+                    if (ch == '\n')
+                      bad_prog(UNTERM_Y_CMD);
+                    if (ch == '\\')
+                      ch = inchar();
+                    if (ch == EOF)
+                      bad_prog(UNTERM_Y_CMD);
+                    cur_cmd->x.translate[*ustring++] = ch;
+                  }
+                free_buffer(b);
 
-		if (len != dest_len)
-                  bad_prog(_(Y_CMD_LEN));
-
-	        for (len = 0; len < YMAP_LENGTH; len++)
-	          translate[len] = len;
-
-                while (dest_len--)
-                  translate[(unsigned char)*ustring++] = *dest_buf++;
-
-	        cur_cmd->x.translate = translate;
+                if (inchar() != slash)
+		  bad_prog(_(Y_CMD_LEN));
+                else if ((ch = in_nonblank()) != EOF && ch != '\n' && ch != ';')
+                  bad_prog(_(EXCESS_JUNK));
 	      }
-
-            if ((ch = in_nonblank()) != EOF && ch != '\n' && ch != ';')
-              bad_prog(_(EXCESS_JUNK));
-
-            free_buffer(b);
-            free_buffer(b2);
 	  }
 	break;
 
@@ -1430,143 +1372,93 @@ bad_prog(why)
 
 /* deal with \X escapes */
 size_t
-normalize_text(buf, len, eat_backwhack)
+normalize_text(buf, len)
   char *buf;
   size_t len;
-  bool eat_backwhack;
 {
   const char *bufend = buf + len;
   char *p = buf;
   char *q = buf;
 
-#ifdef MBS_SUPPORT
-  int mbclen;
-  MBSTATE_T cur_stat;
-  memset(&cur_stat, 0, sizeof(MBSTATE_T));
-#endif
+  /* I'm not certain whether POSIX.2 allows these escapes.
+     Play it safe for now... */
+  if (POSIXLY_CORRECT && !(extended_regexp_flags))
+    return len;
 
   while (p < bufend)
     {
       int c;
 
-      if (*p == '\\' && p+1 < bufend)
-	{
-	  /* I'm not certain whether POSIX.2 allows these escapes.
-	     Play it safe for now... */
-          if (POSIXLY_CORRECT && !(extended_regexp_flags))
-	    switch ( (c = *++p) )
-	      {
-	      case '\n': /*fall through */
-	      case 'n': *q++ = '\n'; p++; continue;
-
-	      default:
-	        /* we just pass the \ up one level for interpretation */
-	        if (!eat_backwhack)
-		  *q++ = '\\';
-	        break;
-	      }
-          else
-	    switch ( (c = *++p) )
-	      {
+      *q = *p++;
+      if (*q == '\\' && p < bufend)
+	switch ( (c = *p++) )
+	  {
 #if defined __STDC__ && __STDC__-0
-	      case 'a': *q++ = '\a'; p++; continue;
+	  case 'a': *q = '\a'; break;
 #else /* Not STDC; we'll just assume ASCII */
-	      case 'a': *q++ = '\007'; p++; continue;
+	  case 'a': *q = '\007'; break;
 #endif
-	      /* case 'b': *q++ = '\b'; p++; continue; --- conflicts with \b RE */
-	      case 'f': *q++ = '\f'; p++; continue;
-	      case '\n': /*fall through */
-	      case 'n': *q++ = '\n'; p++; continue;
-	      case 'r': *q++ = '\r'; p++; continue;
-	      case 't': *q++ = '\t'; p++; continue;
-	      case 'v': *q++ = '\v'; p++; continue;
+	  /* case 'b': *q = '\b'; break; --- conflicts with \b RE */
+	  case 'f': *q = '\f'; break;
+	  case '\n': /*fall through */
+	  case 'n': *q = '\n'; break;
+	  case 'r': *q = '\r'; break;
+	  case 't': *q = '\t'; break;
+	  case 'v': *q = '\v'; break;
 
-	      case 'd': /* decimal byte */
-	        p = convert_number(q, p+1, bufend, 10, 3, 'd');
-	        q++;
-	        continue;
+	  case 'd': /* decimal byte */
+	    p = convert_number(q, p, bufend, 10, 3, 'd');
+	    break;
 
-	      case 'x': /* hexadecimal byte */
-	        p = convert_number(q, p+1, bufend, 16, 2, 'x');
-	        q++;
-	        continue;
+	  case 'x': /* hexadecimal byte */
+	    p = convert_number(q, p, bufend, 16, 2, 'x');
+	    break;
 
 #ifdef REG_PERL
-	      case '0': case '1': case '2': case '3':
-	      case '4': case '5': case '6': case '7':
-	        if ((extended_regexp_flags & REG_PERL) && p+1 < bufend && isdigit(p[1]))
-	          {
-		    p = convert_number(q, p, bufend, 8, 3, *p);
-		    q++;
-	          }
-	        else
-	          {
-		    /* we just pass the \ up one level for interpretation */
-	            if (!eat_backwhack)
-		      *q++ = '\\';
-		    break;
-	          }
-
-	        continue;
-
-	      case 'o': /* octal byte */
-	        if (!(extended_regexp_flags & REG_PERL))
-	          {
-	            p = convert_number(q, p+1, bufend,  8, 3, 'o');
-		    q++;
-	          }
-	        else
-	          {
-	            /* we just pass the \ up one level for interpretation */
-	            if (!eat_backwhack)
-		      *q++ = '\\';
-	            break;
-	          }
-	    
-	        continue;
-#else
-	      case 'o': /* octal byte */
-	        p = convert_number(q, p+1, bufend,  8, 3, 'o');
-	        q++;
-	        continue;
-#endif
-
-	      case 'c':
-	        if (++p < bufend)
-	          {
-		    *q++ = toupper(*p) ^ 0x40;
-		    p++;
-		    continue;
-	          }
-	        else
-	          {
-	            /* we just pass the \ up one level for interpretation */
-	            if (!eat_backwhack)
-		      *q++ = '\\';
-	            break;
-	          }
-
-	      default:
-	        /* we just pass the \ up one level for interpretation */
-	        if (!eat_backwhack)
-		  *q++ = '\\';
-	        break;
+	  case '0': case '1': case '2': case '3':
+	  case '4': case '5': case '6': case '7':
+	    if ((extended_regexp_flags & REG_PERL) &&
+		p < bufend && isdigit(*p))
+	      {
+		p--;
+		p = convert_number(q, p, bufend, 8, 3, *p);
 	      }
-	  }
+	    else
+	      /* we just pass the \ up one level for interpretation */
+	      *++q = p[-1];
 
-#ifdef MBS_SUPPORT
-      mbclen = mbrlen(p, bufend - p, &cur_stat);
-      /* An invalid sequence, or a truncated multibyte character.
-         We treat it as a singlebyte character.  */
-      if (mbclen == (size_t) -1 || mbclen == (size_t) -2 || mbclen == 0)
-        mbclen = 1;
+	    break;
+
+	  case 'o': /* octal byte */
+	    if (!(extended_regexp_flags & REG_PERL))
+	      p = convert_number(q, p, bufend,  8, 3, 'o');
+	    else
+	      /* we just pass the \ up one level for interpretation */
+	      *++q = p[-1];
+	    
+	    break;
 
-      memmove (q, p, mbclen);
-      q += mbclen;
-      p += mbclen;
 #else
-      *q++ = *p++;
+	  case 'o': /* octal byte */
+	    p = convert_number(q, p, bufend,  8, 3, 'o');
+	    break;
 #endif
+
+	  case 'c':
+	    if (p < bufend)
+	      {
+		*q = toupper(*p) ^ 0x40;
+		p++;
+		break;
+	      }
+	    /* else FALLTHROUGH */
+
+	  default:
+	    /* we just pass the \ up one level for interpretation */
+	    *++q = p[-1];
+	    break;
+	  }
+      ++q;
     }
     return (size_t)(q - buf);
 }
@@ -1597,7 +1489,7 @@ compile_string(cur_program, str, len)
   prog.cur = NULL;
   prog.end = NULL;
 
-  first_script = false;
+  first_script = FALSE;
   return ret;
 }
 
@@ -1614,7 +1506,7 @@ compile_file(cur_program, cmdfile)
 
   prog.file = stdin;
   if (cmdfile[0] != '-' || cmdfile[1] != '\0')
-    prog.file = ck_fopen(cmdfile, "rt", true);
+    prog.file = ck_fopen(cmdfile, "rt", TRUE);
 
   cur_input.line = 1;
   cur_input.name = cmdfile;
@@ -1625,7 +1517,7 @@ compile_file(cur_program, cmdfile)
     ck_fclose(prog.file);
   prog.file = NULL;
 
-  first_script = false;
+  first_script = FALSE;
   return ret;
 }
 
@@ -1643,7 +1535,7 @@ check_final_program(program)
   if (blocks)
     {
       /* update info for error reporting: */
-      MEMCPY(&cur_input, &blocks->err_info, sizeof (cur_input));
+      MEMCPY(&cur_input, &blocks->err_info, sizeof cur_input);
       bad_prog(_(EXCESS_OPEN_BRACE));
     }
 
@@ -1681,7 +1573,7 @@ check_final_program(program)
 
   /* There is no longer a need to track file names: */
   {
-    struct output *p;
+    struct fp_list *p;
 
     for (p=file_read; p; p=p->link)
       if (p->name)
@@ -1699,14 +1591,14 @@ check_final_program(program)
   }
 }
 
-/* Rewind all resources which were allocated in this module. */
+/* Release all resources which were allocated in this module. */
 void
 rewind_read_files()
 {
-  struct output *p;
+  struct fp_list *p, *q;
 
   for (p=file_read; p; p=p->link)
-    if (p->fp)
+    if (p->fp && !p->special)
       rewind(p->fp);
 }
 
@@ -1717,11 +1609,11 @@ finish_program(program)
 {
   /* close all files... */
   {
-    struct output *p, *q;
+    struct fp_list *p, *q;
 
     for (p=file_read; p; p=q)
       {
-	if (p->fp)
+	if (p->fp && !p->special)
 	  ck_fclose(p->fp);
 	q = p->link;
 #if 0
@@ -1732,7 +1624,7 @@ finish_program(program)
 
     for (p=file_write; p; p=q)
       {
-	if (p->fp)
+	if (p->fp && !p->special)
 	  ck_fclose(p->fp);
 	q = p->link;
 #if 0
