@@ -54,6 +54,14 @@ extern int errno;
 # include <memory.h>
 #endif
 
+
+#if defined HAVE_WCTYPE_H && defined HAVE_WCHAR_H && defined HAVE_MBRTOWC
+/* We can handle multibyte string.  */
+# include <wchar.h>
+# include <wctype.h>
+# define MBS_SUPPORT
+#endif
+
 #ifndef HAVE_STRCHR
 # define strchr index
 # define strrchr rindex
@@ -72,6 +80,7 @@ extern int errno;
 
 #include <sys/stat.h>
 
+#include "basicdefs.h"
 #include "utils.h"
 #include "sed.h"
 
@@ -82,10 +91,7 @@ struct line {
   char *active;		/* Pointer to non-consumed part of text. */
   size_t length;	/* Length of text (or active, if used). */
   size_t alloc;		/* Allocated space for active. */
-  bool chomped;		/* Was a trailing newline dropped? */
-#ifdef MBS_SUPPORT
-  mbstate_t mbstate;
-#endif
+  flagT chomped;	/* Was a trailing newline dropped? */
 };
 
 /* A queue of text to write out at the end of a cycle
@@ -95,7 +101,7 @@ struct append_queue {
   char *text;
   size_t textlen;
   struct append_queue *next;
-  bool free;
+  flagT free;
 };
 
 /* State information for the input stream. */
@@ -107,7 +113,7 @@ struct input {
   countT bad_count;	/* count of files we failed to open */
   countT line_number;	/* current input line number (over all files) */
 
-  bool (*read_fn) P_((struct input *));	/* read one line */
+  flagT (*read_fn) P_((struct input *));	/* read one line */
   /* If fp is NULL, read_fn better not be one which uses fp;
      in particular, read_always_fail() is recommended. */
 
@@ -115,15 +121,15 @@ struct input {
   const char *in_file_name;
 
   FILE *fp;		/* if NULL, none of the following are valid */
-  bool no_buffering;
+  flagT no_buffering;
 };
 
 
 /* Have we done any replacements lately?  This is used by the `t' command. */
-static bool replaced = false;
+static flagT replaced = FALSE;
 
 /* The current output file (stdout if -i is not being used. */
-static struct output output_file;
+static FILE *output_file;
 
 /* The `current' input line. */
 static struct line line;
@@ -218,28 +224,6 @@ str_append(to, string, length)
     resize_line(to, new_length);
   MEMCPY(to->active + to->length, string, length);
   to->length = new_length;
-
-#ifdef MBS_SUPPORT
-  if (mb_cur_max == 1)
-    return;
-
-  while (length)
-    {
-      int n = MBRLEN (string, length, &to->mbstate);
-
-      /* An invalid sequence is treated like a singlebyte character. */
-      if (n == -1)
-	{
-	  memset (&to->mbstate, 0, sizeof (to->mbstate));
-	  n = 1;
-	}
-
-      if (n > 0)
-	length -= n;
-      else
-	break;
-    }
-#endif
 }
 
 static void str_append_modified P_((struct line *, const char *, size_t,
@@ -257,102 +241,34 @@ str_append_modified(to, string, length, type)
   if (length == 0)
     return;
 
-#ifdef MBS_SUPPORT
-  {
-    MBSTATE_T from_stat;
-
-    if (type == REPL_ASIS)
-      {
-	str_append(to, string, length);
-        return;
-      }
-
-    if (to->alloc - to->length < length * mb_cur_max)
-      resize_line(to, to->length + length * mb_cur_max);
-
-    MEMCPY (&from_stat, &to->mbstate, sizeof(MBSTATE_T));
-    while (length)
-      {
-	wchar_t wc;
-        int n = MBRTOWC (&wc, string, length, &from_stat);
-
-        /* An invalid sequence is treated like a singlebyte character. */
-        if (n == -1)
-          {
-            memset (&to->mbstate, 0, sizeof (from_stat));
-            n = 1;
-          }
-
-        if (n > 0)
-          string += n, length -= n;
-        else
-	  {
-	    /* Incomplete sequence, copy it manually.  */
-	    str_append(to, string, length);
-	    return;
-	  }
-
-	/* Convert the first character specially... */
-        if (type & (REPL_UPPERCASE_FIRST | REPL_LOWERCASE_FIRST))
-	  {
-            if (type & REPL_UPPERCASE_FIRST)
-              wc = towupper(wc);
-            else
-              wc = towlower(wc);
-
-            type &= ~(REPL_LOWERCASE_FIRST | REPL_UPPERCASE_FIRST);
-	    if (type == REPL_ASIS)
-	      {
-		str_append(to, string, length);
-	        return;
-	      }
-          }
-
-        else if (type & REPL_UPPERCASE)
-          wc = towupper(wc);
-        else
-          wc = towlower(wc);
-
-	/* Copy the new wide character to the end of the string. */
-	n = wcrtomb (to->active + to->length, wc, &to->mbstate);
-        to->length += n;
-	if (n == -1)
-	  {
-	    printf ("Invalid multibyte sequence after case conversion!");
-	    abort ();
-	  }
-      }
-  }
-#else
   str_append(to, string, length);
   start = to->active + old_length;
   end = start + length;
 
   /* Now do the required modifications.  First \[lu]... */
-  if (type & REPL_UPPERCASE_FIRST)
+  if (type & repl_uppercase_first)
     {
       *start = toupper(*start);
       start++;
-      type &= ~REPL_UPPERCASE_FIRST;
+      type &= ~repl_uppercase_first;
     }
-  else if (type & REPL_LOWERCASE_FIRST)
+  else if (type & repl_lowercase_first)
     {
       *start = tolower(*start);
       start++;
-      type &= ~REPL_LOWERCASE_FIRST;
+      type &= ~repl_lowercase_first;
     }
 
-  if (type == REPL_ASIS)
+  if (type == repl_asis)
     return;
 
   /* ...and then \[LU] */
-  if (type == REPL_UPPERCASE)
+  if (type == repl_uppercase)
     for (; start != end; start++)
       *start = toupper(*start);
   else
     for (; start != end; start++)
       *start = tolower(*start);
-#endif
 }
 
 /* initialize a "struct line" buffer */
@@ -366,12 +282,7 @@ line_init(buf, initial_size)
   buf->active = buf->text;
   buf->alloc = initial_size;
   buf->length = 0;
-  buf->chomped = true;
-
-#ifdef MBS_SUPPORT
-  memset (&buf->mbstate, 0, sizeof (buf->mbstate));
-#endif
-
+  buf->chomped = TRUE;
 }
 
 /* Copy the contents of the line `from' into the line `to'.
@@ -402,10 +313,6 @@ line_copy(from, to)
   to->length = from->length;
   to->chomped = from->chomped;
   MEMCPY(to->active, from->active, from->length);
-
-#ifdef MBS_SUPPORT
-  MEMCPY(&to->mbstate, &from->mbstate, sizeof (from->mbstate));
-#endif
 }
 
 /* Append the contents of the line `from' to the line `to'. */
@@ -418,10 +325,6 @@ line_append(from, to)
   str_append(to, "\n", 1);
   str_append(to, from->active, from->length);
   to->chomped = from->chomped;
-
-#ifdef MBS_SUPPORT
-  memcpy(&to->mbstate, &from->mbstate, sizeof (from->mbstate));
-#endif
 }
 
 /* Exchange the contents of two "struct line" buffers. */
@@ -440,16 +343,16 @@ line_exchange(a, b)
 
 
 /* dummy function to simplify read_pattern_space() */
-static bool read_always_fail P_((struct input *));
-static bool
+static flagT read_always_fail P_((struct input *));
+static flagT
 read_always_fail(input)
   struct input *input UNUSED;
 {
-  return false;
+  return FALSE;
 }
 
-static bool read_file_line P_((struct input *));
-static bool
+static flagT read_file_line P_((struct input *));
+static flagT
 read_file_line(input)
   struct input *input;
 {
@@ -458,61 +361,38 @@ read_file_line(input)
 
   long result = getline (&b, &blen, input->fp);
 
-  if (result <= 0)
-    return false;
-
   /* Remove the trailing new-line that is left by getline. */
-  else if (b[result - 1] == '\n')
+  if (result > 0 && b[result - 1] == '\n')
     --result;
-
   else
-    line.chomped = false;
+    {
+      /* No trailing new line found. */
+      if (!*input->file_list && !POSIXLY_CORRECT)
+	line.chomped = FALSE;
+
+      if (result <= 0)
+	return FALSE;
+    }
 
   str_append(&line, b, result);
-  return true;
+  return TRUE;
 }
 
 
-static inline void output_missing_newline P_((struct output *));
-static inline void
-output_missing_newline(outf)
-  struct output *outf;
-{
-  if (outf->missing_newline)
-    {
-      ck_fwrite("\n", 1, 1, outf->fp);
-      outf->missing_newline = false;
-    }
-}
-
-static inline void flush_output P_((FILE *));
-static inline void
-flush_output(fp)
-  FILE *fp;
-{
-  if (fp != stdout || unbuffered_output)
-    ck_fflush(fp);
-}
-
-static void output_line P_((const char *, size_t, bool, struct output *));
+static void output_line P_((const char *, size_t, flagT, FILE *));
 static void
-output_line(text, length, nl, outf)
+output_line(text, length, nl, fp)
   const char *text;
   size_t length;
-  bool nl;
-  struct output *outf;
+  flagT nl;
+  FILE *fp;
 {
-  output_missing_newline(outf);
-
   if (length)
-    ck_fwrite(text, 1, length, outf->fp);
-
+    ck_fwrite(text, 1, length, fp);
   if (nl)
-    ck_fwrite("\n", 1, 1, outf->fp);
-  else
-    outf->missing_newline = true;
-
-  flush_output(outf->fp);
+    ck_fwrite("\n", 1, 1, fp);
+  if (fp != stdout || unbuffered_output)
+    ck_fflush(fp);
 }
 
 static struct append_queue *next_append_slot P_((void));
@@ -525,7 +405,7 @@ next_append_slot()
   n->text = NULL;
   n->textlen = 0;
   n->next = NULL;
-  n->free = false;
+  n->free = FALSE;
 
   if (append_tail)
       append_tail->next = n;
@@ -557,12 +437,10 @@ dump_append_queue()
 {
   struct append_queue *p;
 
-  output_missing_newline(&output_file);
   for (p=append_head; p; p=p->next)
     {
       if (p->text)
-        ck_fwrite(p->text, 1, p->textlen, output_file.fp);
-
+	  output_line(p->text, p->textlen, FALSE, output_file);
       if (p->fname)
 	{
 	  char buf[FREAD_BUFFER_SIZE];
@@ -573,17 +451,15 @@ dump_append_queue()
 	     be treated as if it were an empty file, causing no error
 	     condition."  IEEE Std 1003.2-1992
 	     So, don't fail. */
-	  fp = ck_fopen(p->fname, "r", false);
+	  fp = ck_fopen(p->fname, "r", FALSE);
 	  if (fp)
 	    {
 	      while ((cnt = ck_fread(buf, 1, sizeof buf, fp)) > 0)
-		ck_fwrite(buf, 1, cnt, output_file.fp);
+		ck_fwrite(buf, 1, cnt, output_file);
 	      ck_fclose(fp);
 	    }
 	}
     }
-
-  flush_output(output_file.fp);
   release_append_queue();
 }
 
@@ -635,7 +511,7 @@ open_next_file(name, input)
       clearerr(stdin);	/* clear any stale EOF indication */
       input->fp = stdin;
     }
-  else if ( ! (input->fp = ck_fopen(name, "r", false)) )
+  else if ( ! (input->fp = ck_fopen(name, "r", FALSE)) )
     {
       const char *ptr = strerror(errno);
       fprintf(stderr, _("%s: can't read %s: %s\n"), myname, name, ptr);
@@ -662,9 +538,6 @@ open_next_file(name, input)
       output_fd = mkstemp (input->out_file_name);
       free (tmpdir);
 
-      if (output_fd == -1)
-        panic(_("Couldn't open temporary file %s: %s"), input->out_file_name, strerror(errno));
-
 #ifdef HAVE_FCHMOD
       {
         struct stat st;
@@ -673,14 +546,10 @@ open_next_file(name, input)
       }
 #endif
 
-      output_file.fp = fdopen (output_fd, "w");
-      output_file.missing_newline = false;
-
-      if (!output_file.fp)
-        panic(_("Couldn't open temporary file %s: %s"), input->out_file_name, strerror(errno));
+      output_file = fdopen (output_fd, "w");
     }
   else
-    output_file.fp = stdout;
+    output_file = stdout;
 }
 
 
@@ -696,23 +565,24 @@ closedown(input)
   if (input->fp != stdin) /* stdin can be reused on tty and tape devices */
     ck_fclose(input->fp);
 
-  if (in_place_extension && output_file.fp != NULL)
-    {
-      int fd = fileno (output_file.fp);
-      if (strcmp(in_place_extension, "*") != 0)
-        {
-          char *backup_file_name = get_backup_file_name(input->in_file_name);
-          rename (input->in_file_name, backup_file_name);
-          free (backup_file_name);
-        }
+      if (in_place_extension && output_file != NULL)
+	{
+	  if (strcmp(in_place_extension, "*") != 0)
+	    {
+	      char *backup_file_name = get_backup_file_name(input->in_file_name);
+	      rename (input->in_file_name, backup_file_name);
+	      free (backup_file_name);
+	    }
 
-      ck_fclose (output_file.fp);
-      close (fd);
-      rename (input->out_file_name, input->in_file_name);
-      free (input->out_file_name);
-    }
+	  ck_fclose (output_file);
+	  rename (input->out_file_name, input->in_file_name);
+	  free (input->out_file_name);
+	}
 
   input->fp = NULL;
+
+  if (separate_files)
+    rewind_read_files ();
 }
 
 /* Reset range commands so that they are marked as non-matching */
@@ -725,59 +595,51 @@ reset_addresses(vec)
   int n;
 
   for (cur_cmd = vec->v, n = vec->v_length; n--; cur_cmd++)
-    {
-      enum addr_types type = cur_cmd->a1->addr_type;
-      cur_cmd->a1_matched = (type == ADDR_IS_NUM || type == ADDR_IS_NUM_MOD)
-			    && cur_cmd->a1->addr_number == 0;
-    }
+    cur_cmd->a1_matched = FALSE;
 }
 
 /* Read in the next line of input, and store it in the pattern space.
    Return zero if there is nothing left to input. */
-static bool read_pattern_space P_((struct input *, struct vector *, bool));
-static bool
+static flagT read_pattern_space P_((struct input *, struct vector *, flagT));
+static flagT
 read_pattern_space(input, the_program, append)
   struct input *input;
   struct vector *the_program;
-  bool append;
+  flagT append;
 {
   if (append_head) /* redundant test to optimize for common case */
     dump_append_queue();
-  replaced = false;
+  replaced = FALSE;
   if (!append)
     line.length = 0;
-  line.chomped = true;  /* default, until proved otherwise */
+  line.chomped = TRUE;  /* default, until proved otherwise */
 
   while ( ! (*input->read_fn)(input) )
     {
       closedown(input);
 
       if (!*input->file_list)
-	return false;
+	{
+	  line.chomped = FALSE;
+	  return FALSE;
+	}
 
       if (separate_files)
 	{
 	  input->line_number = 0;
 	  reset_addresses (the_program);
-	  rewind_read_files ();
-
-	  /* If doing in-place editing, we will never append the
-	     new-line to this file; but if the output goes to stdout,
-	     we might still have to output the missing new-line.  */
-	  if (in_place_extension)
-	    output_file.missing_newline = false;
 	}
 
       open_next_file (*input->file_list++, input);
     }
 
   ++input->line_number;
-  return true;
+  return TRUE;
 }
 
 
-static bool last_file_with_data_p P_((struct input *));
-static bool
+static flagT last_file_with_data_p P_((struct input *));
+static flagT
 last_file_with_data_p(input)
   struct input *input;
 {
@@ -787,29 +649,29 @@ last_file_with_data_p(input)
 
       closedown(input);
       if (!*input->file_list)
-	return true;
+	return TRUE;
       open_next_file(*input->file_list++, input);
       if (input->fp)
 	{
 	  if ((ch = getc(input->fp)) != EOF)
 	    {
 	      ungetc(ch, input->fp);
-	      return false;
+	      return FALSE;
 	    }
 	}
     }
 }
 
 /* Determine if we match the `$' address. */
-static bool test_eof P_((struct input *));
-static bool
+static flagT test_eof P_((struct input *));
+static flagT
 test_eof(input)
   struct input *input;
 {
   int ch;
 
   if (buffer.length)
-    return false;
+    return FALSE;
   if (!input->fp)
     return separate_files || last_file_with_data_p(input);
   if (feof(input->fp))
@@ -817,84 +679,90 @@ test_eof(input)
   if ((ch = getc(input->fp)) == EOF)
     return separate_files || last_file_with_data_p(input);
   ungetc(ch, input->fp);
-  return false;
+  return FALSE;
 }
 
 /* Return non-zero if the current line matches the address
    pointed to by `addr'. */
-static bool match_an_address_p P_((struct addr *, struct input *));
-static bool
+static flagT match_an_address_p P_((struct addr *, struct input *));
+static flagT
 match_an_address_p(addr, input)
   struct addr *addr;
   struct input *input;
 {
   switch (addr->addr_type)
     {
-    case ADDR_IS_NULL:
-      return true;
+    case addr_is_null:
+      return TRUE;
 
-    case ADDR_IS_REGEX:
+    case addr_is_regex:
       return match_regex(addr->addr_regex, line.active, line.length, 0, NULL, 0);
 
-    case ADDR_IS_NUM:
+    case addr_is_num:
       return (addr->addr_number == input->line_number);
 
-    case ADDR_IS_NUM_MOD:
-      return (input->line_number >= addr->addr_number
-	      && ((input->line_number - addr->addr_number) % addr->addr_step) == 0);
+    case addr_is_num_mod:
+      if (addr->addr_number < addr->addr_step)
+	return (addr->addr_number == input->line_number%addr->addr_step);
+      /* addr_number >= step implies we have an extra initial skip */
+      if (input->line_number < addr->addr_number)
+	return FALSE;
+      /* normalize */
+      addr->addr_number %= addr->addr_step;
+      return (addr->addr_number == 0);
 
-    case ADDR_IS_NUM2:
-    case ADDR_IS_STEP:
-    case ADDR_IS_STEP_MOD:
+    case addr_is_num2:
+    case addr_is_step:
+    case addr_is_step_mod:
       /* reminder: these are only meaningful for a2 addresses */
       /* a2->addr_number needs to be recomputed each time a1 address
          matches for the step and step_mod types */
       return (addr->addr_number <= input->line_number);
 
-    case ADDR_IS_LAST:
+    case addr_is_last:
       return test_eof(input);
 
     default:
       panic(_("INTERNAL ERROR: bad address type"));
     }
   /*NOTREACHED*/
-  return false;
+  return FALSE;
 }
 
 /* return non-zero if current address is valid for cmd */
-static bool match_address_p P_((struct sed_cmd *, struct input *));
-static bool
+static flagT match_address_p P_((struct sed_cmd *, struct input *));
+static flagT
 match_address_p(cmd, input)
   struct sed_cmd *cmd;
   struct input *input;
 {
-  bool addr_matched = cmd->a1_matched;
+  flagT addr_matched = cmd->a1_matched;
 
   if (addr_matched)
     {
       if (match_an_address_p(cmd->a2, input))
-	cmd->a1_matched = false;
+	cmd->a1_matched = FALSE;
     }
   else if (!cmd->a1 || match_an_address_p(cmd->a1, input))
     {
-      addr_matched = true;
+      addr_matched = TRUE;
       if (cmd->a2)
 	{
-	  cmd->a1_matched = true;
+	  cmd->a1_matched = TRUE;
 	  switch (cmd->a2->addr_type)
 	    {
-	    case ADDR_IS_REGEX:
+	    case addr_is_regex:
 	      break;
-	    case ADDR_IS_STEP:
+	    case addr_is_step:
 	      cmd->a2->addr_number = input->line_number + cmd->a2->addr_step;
 	      break;
-	    case ADDR_IS_STEP_MOD:
+	    case addr_is_step_mod:
 	      cmd->a2->addr_number = input->line_number + cmd->a2->addr_step
 				     - (input->line_number%cmd->a2->addr_step);
 	      break;
 	    default:
 	      if (match_an_address_p(cmd->a2, input))
-		cmd->a1_matched = false;
+		cmd->a1_matched = FALSE;
 	      break;
 	    }
 	}
@@ -916,9 +784,7 @@ do_list(line_len)
   char obuf[180];	/* just in case we encounter a 512-bit char (;-) */
   char *o;
   size_t olen;
-  FILE *fp = output_file.fp;
 
-  output_missing_newline(&output_file);
   for (; len--; ++p) {
       o = obuf;
       
@@ -955,61 +821,13 @@ do_list(line_len)
       }
       olen = o - obuf;
       if (width+olen >= line_len && line_len > 0) {
-	  ck_fwrite("\\\n", 1, 2, fp);
+	  ck_fwrite("\\\n", 1, 2, output_file);
 	  width = 0;
       }
-      ck_fwrite(obuf, 1, olen, fp);
+      ck_fwrite(obuf, 1, olen, output_file);
       width += olen;
   }
-  ck_fwrite("$\n", 1, 2, fp);
-  flush_output (fp);
-}
-
-
-static enum replacement_types append_replacement P_((struct line *, struct replacement *,
-						     struct re_registers *,
-						     enum replacement_types));
-static enum replacement_types
-append_replacement (buf, p, regs, repl_mod)
-  struct line *buf;
-  struct replacement *p;
-  struct re_registers *regs;
-  enum replacement_types repl_mod;
-{
-  for (; p; p=p->next)
-    {
-      int i = p->subst_id;
-      enum replacement_types curr_type;
-
-      /* Apply a \[lu] modifier that was given earlier, but which we
-         have not had yet the occasion to apply.  But don't do it
-         if this replacement has a modifier of its own. */
-      curr_type = (p->repl_type & REPL_MODIFIERS)
-        ? p->repl_type
-        : p->repl_type | repl_mod;
-
-      repl_mod = 0;
-      if (p->prefix_length)
-        {
-          str_append_modified(buf, p->prefix, p->prefix_length,
-    			      curr_type);
-          curr_type &= ~REPL_MODIFIERS;
-        }
-
-      if (0 <= i)
-        if (regs->end[i] == regs->start[i] && p->repl_type & REPL_MODIFIERS)
-          /* Save this modifier, we shall apply it later.
-	     e.g. in s/()([a-z])/\u\1\2/
-	     the \u modifier is applied to \2, not \1 */
-	  repl_mod = curr_type & REPL_MODIFIERS;
-
-	else
-	  str_append_modified(buf, line.active + regs->start[i],
-			      CAST(size_t)(regs->end[i] - regs->start[i]),
-			      curr_type);
-    }
-
-  return repl_mod;
+  ck_fwrite("$\n", 1, 2, output_file);
 }
 
 static void do_subst P_((struct subst *));
@@ -1020,7 +838,6 @@ do_subst(sub)
   size_t start = 0;	/* where to start scan for (next) match in LINE */
   size_t last_end = 0;  /* where did the last successful match end in LINE */
   countT count = 0;	/* number of matches found */
-  bool again = true;
 
 #define MAX_BACKREFERENCES 10
   static struct re_registers regs;
@@ -1039,24 +856,25 @@ do_subst(sub)
     if (regs.start[0] == 0 && !sub->global)
       {
 	/* We found a match, set the `replaced' flag. */
-	replaced = true;
+	replaced = TRUE;
 
 	line.active += regs.end[0];
 	line.length -= regs.end[0];
 	line.alloc -= regs.end[0];
-	goto post_subst;
+	goto did_subst;
       }
     else if (regs.end[0] == line.length)
       {
 	/* We found a match, set the `replaced' flag. */
-	replaced = true;
+	replaced = TRUE;
 
 	line.length = regs.start[0];
-	goto post_subst;
+	goto did_subst;
       }
 
   do
     {
+      struct replacement *p;
       enum replacement_types repl_mod = 0;
 
       size_t offset = regs.start[0];
@@ -1075,42 +893,74 @@ do_subst(sub)
          s,a*,x,g, gives `xbxxcx' instead of xbxcx.  This behavior is
          unacceptable because it is not consistently applied (for
          example, `baaaa' gives `xbx', not `xbxx'). */
-      if ((matched > 0 || count == 0 || offset > last_end)
-	  && ++count >= sub->numb)
-        {
-          /* We found a match, set the `replaced' flag. */
-          replaced = true;
-
-          /* Now expand the replacement string into the output string. */
-          repl_mod = append_replacement (&s_accum, sub->replacement, &regs, repl_mod);
-	  again = sub->global;
-        }
-      else
+      if ((count > 0 && offset == last_end && matched == 0)
+	  || ++count < sub->numb)
 	{
-          /* The match was not replaced.  Copy the text until its
-             end; if it was vacuous, skip over one character and
-	     add that character to the output.  */
-	  if (matched == 0)
-	    {
-	      if (start < line.length)
-	        matched = 1;
-	      else
-	        break;
-	    }
+	  /* If the match was vacuous, skip ahead one character
+	   * anyway.
+	   */
+	  if (matched == 0 && offset < line.length)
+	    matched = 1;
 
 	  str_append(&s_accum, line.active + offset, matched);
-        }
+	  start = offset + matched;
+	  continue;
+	}
+
+      /* We found a match, set the `replaced' flag. */
+      replaced = TRUE;
 
-      /* Start after the match.  last_end is the real end of the matched
-	 substring, excluding characters that were skipped in case the RE
-	 matched the empty string.  */
-      start = offset + matched;
       last_end = regs.end[0];
+
+      /* Now expand the replacement string into the output string. */
+      for (p=sub->replacement; p; p=p->next)
+	{
+	  int i = p->subst_id;
+	  enum replacement_types curr_type;
+
+	  /* Apply a \[lu] modifier that was given earlier, but which we
+	     have not had yet the occasion to apply.  But don't do it
+	     if this replacement has a modifier of its own. */
+	  curr_type = (p->repl_type & repl_modifiers)
+	    ? p->repl_type
+	    : p->repl_type | repl_mod;
+
+	  repl_mod = 0;
+	  if (p->prefix_length)
+	    {
+	      str_append_modified(&s_accum, p->prefix, p->prefix_length,
+				  curr_type);
+	      curr_type &= ~repl_modifiers;
+	    }
+
+	  if (0 <= i)
+	    if (regs.end[i] == regs.start[i] && p->repl_type & repl_modifiers)
+	      /* Save this modifier, we shall apply it later.
+		 e.g. in s/()([a-z])/\u\1\2/
+		 the \u modifier is applied to \2, not \1 */
+	      repl_mod = curr_type & repl_modifiers;
+
+	    else
+	      str_append_modified(&s_accum, line.active + regs.start[i],
+				  CAST(size_t)(regs.end[i] - regs.start[i]),
+				  curr_type);
+	}
+
+      start = regs.end[0];
+      if (!sub->global || start == line.length)
+	break;
+
+      /* If the match was vacuous, skip over one character
+       * and add that character to the output.
+       */
+      if (regs.start[0] == regs.end[0])
+	{
+	  str_append(&s_accum, line.active + offset, 1);
+	  ++start;
+	}
     }
-  while (again
-	 && start <= line.length
-	 && match_regex(sub->regx, line.active, line.length, start,
-			&regs, MAX_BACKREFERENCES));
+  while (match_regex(sub->regx, line.active, line.length, start,
+		     &regs, MAX_BACKREFERENCES));
 
   /* Copy stuff to the right of the last match into the output string. */
   if (start < line.length)
@@ -1120,14 +970,11 @@ do_subst(sub)
   /* Exchange line and s_accum.  This can be much cheaper
      than copying s_accum.active into line.text (for huge lines). */
   line_exchange(&line, &s_accum);
+ did_subst:
   
   /* Finish up. */
-  if (count < sub->numb)
-    return;
-
- post_subst:
   if (sub->print & 1)
-    output_line(line.active, line.length, line.chomped, &output_file);
+    output_line(line.active, line.length, line.chomped, output_file);
   
   if (sub->eval) 
     {
@@ -1163,9 +1010,9 @@ do_subst(sub)
     } 
   
   if (sub->print & 2)
-    output_line(line.active, line.length, line.chomped, &output_file);
-  if (sub->outf)
-    output_line(line.active, line.length, line.chomped, sub->outf);
+    output_line(line.active, line.length, line.chomped, output_file);
+  if (sub->fp)
+    output_line(line.active, line.length, line.chomped, sub->fp);
 }
 
 #ifdef EXPERIMENTAL_DASH_N_OPTIMIZATION
@@ -1261,20 +1108,24 @@ execute_program(vec, input)
 	    case 'c':
 	      if (!cur_cmd->a1_matched)
 		output_line(cur_cmd->x.cmd_txt.text,
-			    cur_cmd->x.cmd_txt.text_length - 1, true,
-			    &output_file);
+			    cur_cmd->x.cmd_txt.text_length, FALSE, output_file);
 	      /* POSIX.2 is silent about c starting a new cycle,
 		 but it seems to be expected (and make sense). */
 	      /* Fall Through */
 	    case 'd':
+	      line.length = 0;
+	      line.chomped = FALSE;
 	      return -1;
 
 	    case 'D':
 	      {
 		char *p = memchr(line.active, '\n', line.length);
 		if (!p)
-		  return -1;
-
+		  {
+		    line.length = 0;
+		    line.chomped = FALSE;
+		    return -1;
+		  }
 		++p;
 		line.alloc -= p - line.active;
 		line.length -= p - line.active;
@@ -1302,7 +1153,6 @@ execute_program(vec, input)
 		{
 		  cur_cmd->x.cmd_txt.text[cmd_length - 1] = 0;
 		  pipe = popen(cur_cmd->x.cmd_txt.text, "r");
-                  output_missing_newline(&output_file);
 		}
 
 	      if (pipe != NULL) 
@@ -1315,7 +1165,7 @@ execute_program(vec, input)
 			if (!cmd_length)
 			  str_append(&s_accum, buf, n);
 			else
-			  ck_fwrite(buf, 1, n, output_file.fp);
+			  output_line(buf, n, FALSE, output_file);
 		    }
 		  
 		  pclose (pipe);
@@ -1331,9 +1181,6 @@ execute_program(vec, input)
 			 (for huge lines). */
 		      line_exchange(&line, &s_accum);
 		    }
-                  else
-                    flush_output(output_file.fp);
-
 		}
 	      else
 		panic(_("error in subprocess"));
@@ -1361,8 +1208,7 @@ execute_program(vec, input)
 
 	    case 'i':
 	      output_line(cur_cmd->x.cmd_txt.text,
-			  cur_cmd->x.cmd_txt.text_length - 1,
-			  true, &output_file);
+			  cur_cmd->x.cmd_txt.text_length, FALSE, output_file);
 	      break;
 
 	    case 'l':
@@ -1372,45 +1218,44 @@ execute_program(vec, input)
 	      break;
 
 	    case 'L':
-              output_missing_newline(&output_file);
 	      fmt(line.active, line.active + line.length,
 		  cur_cmd->x.int_arg == -1
 		  ? lcmd_out_line_len
 		  : cur_cmd->x.int_arg,
-		  output_file.fp);
-              flush_output(output_file.fp);
+		  output_file);
 	      break;
 
 	    case 'n':
 	      if (!no_default_output)
-		output_line(line.active, line.length, line.chomped, &output_file);
-	      if (test_eof(input) || !read_pattern_space(input, vec, false))
+		output_line(line.active, line.length, line.chomped, output_file);
+	      if (test_eof(input) || !read_pattern_space(input, vec, FALSE))
 		return -1;
 	      break;
 
 	    case 'N':
 	      str_append(&line, "\n", 1);
-	      if (test_eof(input) || !read_pattern_space(input, vec, true))
+	      if (test_eof(input) || !read_pattern_space(input, vec, TRUE))
 		return -1;
 	      break;
 
 	    case 'p':
-	      output_line(line.active, line.length, line.chomped, &output_file);
+	      output_line(line.active, line.length, line.chomped, output_file);
 	      break;
 
 	    case 'P':
 	      {
 		char *p = memchr(line.active, '\n', line.length);
 		output_line(line.active, p ? p - line.active : line.length,
-			    p ? true : line.chomped, &output_file);
+			    p ? 1 : line.chomped, output_file);
 	      }
 	      break;
 
-            case 'q':
-              if (!no_default_output)
-                output_line(line.active, line.length, line.chomped, &output_file);
-
 	    case 'Q':
+	      line.length = 0;
+	      line.chomped = FALSE;
+	      /* Fall through */
+
+	    case 'q':
 	      return cur_cmd->x.int_arg == -1 ? 0 : cur_cmd->x.int_arg;
 
 	    case 'r':
@@ -1434,7 +1279,7 @@ execute_program(vec, input)
 		  if (result != EOF)
 		    {
 		      aq = next_append_slot();
-		      aq->free = true;
+		      aq->free = TRUE;
 		      aq->text = text;
 		      aq->textlen = result;
 		    }
@@ -1448,7 +1293,7 @@ execute_program(vec, input)
 	    case 't':
 	      if (replaced)
 		{
-		  replaced = false;
+		  replaced = FALSE;
 		  cur_cmd = vec->v + cur_cmd->x.jump_index;
 		  continue;
 		}
@@ -1461,13 +1306,13 @@ execute_program(vec, input)
 		  continue;
 		}
 	      else
-		replaced = false;
+		replaced = FALSE;
 	      break;
 
 	    case 'w':
 	      if (cur_cmd->x.fp)
 		output_line(line.active, line.length,
-			    line.chomped, cur_cmd->x.outf);
+			    line.chomped, cur_cmd->x.fp);
 	      break;
 
 	    case 'W':
@@ -1475,7 +1320,7 @@ execute_program(vec, input)
 	        {
 		  char *p = memchr(line.active, '\n', line.length);
 		  output_line(line.active, p ? p - line.active : line.length,
-			      p ? true : line.chomped, cur_cmd->x.outf);
+			      p ? 1 : line.chomped, cur_cmd->x.fp);
 	        }
 	      break;
 
@@ -1485,18 +1330,18 @@ execute_program(vec, input)
 
 	    case 'y':
 	      {
-#ifdef MBS_SUPPORT
-               if (mb_cur_max > 1)
+#if defined MBS_SUPPORT && !defined REG_PERL
+               if (MB_CUR_MAX > 1)
                  {
                    int idx, prev_idx; /* index in the input line.  */
                    char **trans;
-                   mbstate_t mbstate;
-                   memset(&mbstate, 0, sizeof(mbstate_t));
+                   mbstate_t cur_stat;
+                   memset(&cur_stat, 0, sizeof(mbstate_t));
                    for (idx = 0; idx < line.length;)
                      {
                        int mbclen, i;
-                       mbclen = MBRLEN (line.active + idx, line.length - idx,
-                                          &mbstate);
+                       mbclen = mbrlen(line.active + idx, line.length - idx,
+                                       &cur_stat);
                        /* An invalid sequence, or a truncated multibyte
                           character.  We treat it as a singlebyte character.
                        */
@@ -1508,9 +1353,10 @@ execute_program(vec, input)
                        /* `i' indicate i-th translate pair.  */
                        for (i = 0; trans[2*i] != NULL; i++)
                          {
-                           if (strncmp(line.active + idx, trans[2*i], mbclen) == 0)
+                           if (strncmp(line.active + idx, trans[2*i], mbclen)
+                               == 0)
                              {
-                               bool move_remain_buffer = false;
+                               flagT move_remain_buffer = FALSE;
                                int trans_len = strlen(trans[2*i+1]);
 
                                if (mbclen < trans_len)
@@ -1523,12 +1369,12 @@ execute_program(vec, input)
                                        /* And we must resize the buffer.  */
                                        resize_line(&line, new_len);
                                      }
-                                   move_remain_buffer = true;
+                                   move_remain_buffer = TRUE;
                                  }
                                else if (mbclen > trans_len)
                                  {
                                    /* We must truncate the line buffer.  */
-                                   move_remain_buffer = true;
+                                   move_remain_buffer = TRUE;
                                  }
                                prev_idx = idx;
                                if (move_remain_buffer)
@@ -1564,10 +1410,8 @@ execute_program(vec, input)
 	      break;
 
 	    case '=':
-              output_missing_newline(&output_file);
-              fprintf(output_file.fp, "%lu\n",
-                      CAST(unsigned long)input->line_number);
-              flush_output(output_file.fp);
+	      fprintf(output_file, "%lu\n",
+		      CAST(unsigned long)input->line_number);
 	      break;
 
 	    default:
@@ -1577,7 +1421,7 @@ execute_program(vec, input)
 
 #ifdef EXPERIMENTAL_DASH_N_OPTIMIZATION
       /* If our top-level program consists solely of commands with
-       * ADDR_IS_NUM addresses then once we past the last mentioned
+       * addr_is_num addresses then once we past the last mentioned
        * line we should be able to quit if no_default_output is true,
        * or otherwise quickly copy input to output.  Now whether this
        * optimization is a win or not depends on how cheaply we can
@@ -1589,16 +1433,16 @@ execute_program(vec, input)
        *
        * Don't use this when in-place editing is active, because line
        * numbers restart each time then. */
-      else if (!separate_files)
+      else if (output_file == stdout)
 	{
 	  /* can we ever match again? */
-	  if (cur_cmd->a1->addr_type == ADDR_IS_NUM &&
+	  if (cur_cmd->a1->addr_type == addr_is_num &&
 	      ((input->line_number < cur_cmd->a1->addr_number)
 	       != !cur_cmd->addr_bang))
 	    {
 	      /* skip all this next time */
-	      cur_cmd->a1->addr_type = ADDR_IS_NULL;
-	      cur_cmd->addr_bang = true;
+	      cur_cmd->a1->addr_type = addr_is_null;
+	      cur_cmd->addr_bang = TRUE;
 
 	      /* can we make an optimization? */
 	      if (cur_cmd->cmd == 'b' || cur_cmd->cmd == 't'
@@ -1623,9 +1467,6 @@ execute_program(vec, input)
       /* this is buried down here so that a "continue" statement can skip it */
       ++cur_cmd;
     }
-
-    if (!no_default_output)
-      output_line(line.active, line.length, line.chomped, &output_file);
     return -1;
 }
 
@@ -1658,9 +1499,11 @@ process_files(the_program, argv)
   input.fp = NULL;
 
   status = EXIT_SUCCESS;
-  while (read_pattern_space(&input, the_program, false))
+  while (read_pattern_space(&input, the_program, FALSE))
     {
       status = execute_program(the_program, &input);
+      if (!no_default_output)
+	output_line(line.active, line.length, line.chomped, output_file);
       if (status == -1)
 	status = EXIT_SUCCESS;
       else
