@@ -22,7 +22,6 @@
 
 #include "config.h"
 
-
 #include <stdio.h>
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
@@ -49,6 +48,7 @@
 # include <locale.h>
 #endif
 #include "getopt.h"
+#include "basicdefs.h"
 #include "utils.h"
 #include "sed.h"
 
@@ -67,19 +67,19 @@
 int extended_regexp_flags = 0;
 
 /* If set, fflush(stdout) on every line output. */
-bool unbuffered_output = false;
+flagT unbuffered_output = FALSE;
 
 /* If set, don't write out the line unless explicitly told to */
-bool no_default_output = false;
+flagT no_default_output = FALSE;
 
 /* If set, reset line counts on every new file. */
-bool separate_files = false;
+flagT separate_files = FALSE;
 
 /* How do we edit files in-place? (we don't if NULL) */
 char *in_place_extension = NULL;
 
 /* Do we need to be pedantically POSIX compliant? */
-bool POSIXLY_CORRECT;
+flagT POSIXLY_CORRECT;
 
 /* How long should the `l' command's output line be? */
 countT lcmd_out_line_len = 70;
@@ -102,35 +102,33 @@ usage(status)
 
   fprintf(out, _("\
 Usage: %s [OPTION]... {script-only-if-no-other-script} [input-file]...\n\
-\n"), myname);
-
-  fprintf(out, _("-n, --quiet, --silent\n\
-                 suppress automatic printing of pattern space\n"));
-  fprintf(out, _("  -e script, --expression=script\n\
-                 add the script to the commands to be executed\n"));
-  fprintf(out, _("  -f script-file, --file=script-file\n\
-                 add the contents of script-file to the commands to be executed\n"));
-  fprintf(out, _("  -i[SUFFIX], --in-place[=SUFFIX]\n\
-                 edit files in place (makes backup if extension supplied)\n"));
-  fprintf(out, _("  -l N, --line-length=N\n\
-                 specify the desired line-wrap length for the `l' command\n"));
-  fprintf(out, _("  -r, --regexp-extended\n\
-                 use extended regular expressions in the script.\n"));
-  fprintf(out, PERL_HELP);
-  fprintf(out, _("  -s, --separate\n\
+\n\
+  -n, --quiet, --silent\n\
+                 suppress automatic printing of pattern space\n\
+  -e script, --expression=script\n\
+                 add the script to the commands to be executed\n\
+  -f script-file, --file=script-file\n\
+                 add the contents of script-file to the commands to be executed\n\
+  -i[suffix], --in-place[=suffix]\n\
+                 edit files in place (makes backup if extension supplied)\n\
+  -l N, --line-length=N\n\
+                 specify the desired line-wrap length for the `l' command\n\
+  -r, --regexp-extended\n\
+                 use extended regular expressions in the script.\n%s\
+  -s, --separate\n\
                  consider files as separate rather than as a single continuous\n\
-                 long stream.\n"));
-  fprintf(out, _("  -u, --unbuffered\n\
+                 long stream.\n\
+  -u, --unbuffered\n\
                  load minimal amounts of data from the input files and flush\n\
-                 the output buffers more often\n"));
-  fprintf(out, _("      --help     display this help and exit\n"));
-  fprintf(out, _("      --version  output version information and exit\n"));
-  fprintf(out, _("\n\
+                 the output buffers more often\n\
+      --help     display this help and exit\n\
+  -V, --version  output version information and exit\n\
+\n\
 If no -e, --expression, -f, or --file option is given, then the first\n\
 non-option argument is taken as the sed script to interpret.  All\n\
 remaining arguments are names of input files; if no input files are\n\
 specified, then the standard input is read.\n\
-\n"));
+\n"), myname, PERL_HELP);
   fprintf(out, _("E-mail bug reports to: %s .\n\
 Be sure to include the word ``%s'' somewhere in the ``Subject:'' field.\n"),
 	  BUG_ADDRESS, PACKAGE);
@@ -145,9 +143,9 @@ main(argc, argv)
   char **argv;
 {
 #ifdef REG_PERL
-#define SHORTOPTS "snrRue:f:l:i::V:"
+#define SHORTOPTS "shnrRuVe:f:l:i::"
 #else
-#define SHORTOPTS "snrue:f:l:i::V:"
+#define SHORTOPTS "shnruVe:f:l:i::"
 #endif
 
   static struct option longopts[] = {
@@ -163,27 +161,24 @@ main(argc, argv)
     {"silent", 0, NULL, 'n'},
     {"separate", 0, NULL, 's'},
     {"unbuffered", 0, NULL, 'u'},
-    {"version", 0, NULL, 'v'},
+    {"version", 0, NULL, 'V'},
     {"help", 0, NULL, 'h'},
     {NULL, 0, NULL, 0}
   };
 
   int opt;
   int return_code;
-  const char *cols = getenv("COLS");
 
-  initialize_main (&argc, &argv);
+  initialize_main(&argc, &argv);
+
+#if ENABLE_NLS
 #if HAVE_SETLOCALE
   /* Set locale according to user's wishes.  */
   setlocale (LC_ALL, "");
 #endif
-  initialize_mbcs ();
-
-#if ENABLE_NLS
 
   /* Tell program which translations to use and where to find.  */
   bindtextdomain (PACKAGE, LOCALEDIR);
-  textdomain (PACKAGE);
 #endif
 
   POSIXLY_CORRECT = (getenv("POSIXLY_CORRECT") != NULL);
@@ -192,12 +187,15 @@ main(argc, argv)
      the baseline setting of `lcmd_out_line_len'.  The "-1"
      is to avoid gratuitous auto-line-wrap on ttys.
    */
-  if (cols)
-    {
-      countT t = ATOI(cols);
-      if (t > 1)
-	lcmd_out_line_len = t-1;
-    }
+  {
+    const char *p = getenv("COLS");
+    if (p)
+      {
+	countT t = ATOI(p);
+	if (1 < t)
+	  lcmd_out_line_len = t-1;
+      }
+  }
 
   myname = *argv;
   while ((opt = getopt_long(argc, argv, SHORTOPTS, longopts, NULL)) != EOF)
@@ -205,7 +203,7 @@ main(argc, argv)
       switch (opt)
 	{
 	case 'n':
-	  no_default_output = true;
+	  no_default_output = TRUE;
 	  break;
 	case 'e':
 	  the_program = compile_string(the_program, optarg, strlen(optarg));
@@ -215,21 +213,17 @@ main(argc, argv)
 	  break;
 
 	case 'i':
-	  separate_files = true;
+	  separate_files = TRUE;
 	  if (optarg == NULL)
-	    /* use no backups */
-	    in_place_extension = ck_strdup ("*");
-
+	    in_place_extension = "*";
 	  else if (strchr(optarg, '*') == NULL)
-	    in_place_extension = ck_strdup(optarg);
-
-	  else
 	    {
-	      in_place_extension = MALLOC (strlen(optarg) + 2, char);
+	      in_place_extension = xmalloc (strlen(optarg) + 2);
 	      in_place_extension[0] = '*';
 	      strcpy (in_place_extension + 1, optarg);
 	    }
-
+	  else
+	    in_place_extension = ck_strdup(optarg);
 	  break;
 
 	case 'l':
@@ -250,17 +244,17 @@ main(argc, argv)
 #endif
 
 	case 's':
-	  separate_files = true;
+	  separate_files = TRUE;
 	  break;
 
 	case 'u':
-	  unbuffered_output = true;
+	  unbuffered_output = TRUE;
 	  break;
 
-	case 'v':
+	case 'V':
 #ifdef REG_PERL
 	  fprintf(stdout, _("super-sed version %s\n"), VERSION);
-	  fprintf(stdout, _("based on GNU sed version %s\n\n"), SED_FEATURE_VERSION);
+	  fprintf(stdout, _("based on GNU sed version 3.02.80\n\n"));
 #else
 	  fprintf(stdout, _("GNU sed version %s\n"), VERSION);
 #endif
